#include <assert.h>
#include <err.h>
#include <getopt.h>
#include <mysql.h>
#include <stdbool.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#define MAXDATALEN (1024 * 64)
#define MYSQL_PREFETCH_ROWS 100
#define MAX_COLUMN_COUNT 20

struct column {
        char buf[MAXDATALEN];
        my_bool is_null;
        unsigned long len;
        my_bool is_error;
};

static struct option options[] = {
        { "command", required_argument, NULL, 'c' },
        { "help", no_argument, NULL, 'h' },
        { "host", required_argument, NULL, 'H' },
        { "port", required_argument, NULL, 'P' },
        { "user", required_argument, NULL, 'u' },
        { "database", required_argument, NULL, 'D' },
        { "silent", no_argument, NULL, 's' },
};

static const char *default_query =
        "SELECT "
        "  amount, "
        "  focus, "
        "  executed_on, "
        "  id, "
        "  method, "
        "  price_plan_id, "
        "  source, "
        "  status, "
        "  subscription_id, "
        "  type "
        "FROM foo";

static void print_usage(FILE *f, const char *prog)
{
        fprintf(f,
                "Usage: %s [options]\n"
                "  -h, --help          Print this message and exit.\n"
                "  -H, --host=HOST     Connect to address HOST.\n"
                "  -P, --port=N        Connect to port N.\n"
                "  -u, --user=USER     Connect as user USER.\n"
                "  -D, --database=DB   Use database DB.\n"
                "  -c, --command=SQL   Execute SQL command.\n"
                "  -s, --silent        Only print row count.\n",
                prog);
}

int main(int argc, char **argv)
{
        MYSQL *conn = NULL;
        MYSQL_STMT *stmt = NULL;
        MYSQL_BIND binds[MAX_COLUMN_COUNT] = { 0 };
        int rc;
        unsigned count = 0;
	unsigned long prefetch_rows = MYSQL_PREFETCH_ROWS;
	unsigned long type = (unsigned long)CURSOR_TYPE_READ_ONLY;
        bool silent = false;
        char c;

        const char *query = default_query;
        const char *host = "127.0.0.1";
        unsigned port = 3306;
        const char *user = "root";
        const char *database = "test";
        const char *password;

        struct column columns[MAX_COLUMN_COUNT] = { { 0 } };
        unsigned column_count;

        while ((c = getopt_long(argc, (char * const *)argv, "c:hH:P:u:D:s", options, NULL)) != -1) {
                switch (c) {
                case 'c': query = optarg; break;
                case 'h': print_usage(stdout, argv[0]); return 0;
                case 'H': host = optarg; break;
                case 'P': port = atoi(optarg); break;
                case 'u': user = optarg; break;
                case 'D': database = optarg; break;
                case 's': silent = true; break;
                default: print_usage(stderr, argv[0]); return 1;
                };
        }

        if ((password = getenv("MYSQLPASSWORD")) == NULL) {
                fprintf(stderr, "Please specify MYSQLPASSWORD environmental variable.\n");
                return 1;
        }

        if ((conn = mysql_init(NULL)) == NULL)
                errx(1, "mysql_init");

        if (mysql_real_connect(conn, host, user, password, database, port, NULL, 0) == NULL)
                errx(1, "mysql_real_connect");

        if ((stmt = mysql_stmt_init(conn)) == NULL)
                errx(1, "mysql_stmt_init");

        if (mysql_stmt_prepare(stmt, query, strlen(query)) != 0)
                errx(1, "mysql_stmt_prepare: %s", mysql_error(conn));

        mysql_stmt_attr_set(stmt, STMT_ATTR_CURSOR_TYPE, (void*)&type);
	mysql_stmt_attr_set(stmt, STMT_ATTR_PREFETCH_ROWS, (void*)&prefetch_rows);

        column_count = mysql_stmt_field_count(stmt);
        for (int i = 0; i < column_count; i++) {
                binds[i].is_null = &columns[i].is_null;
                binds[i].length = &columns[i].len;
                binds[i].error = &columns[i].is_error;
                binds[i].buffer_type = MYSQL_TYPE_VAR_STRING;
                binds[i].buffer = &columns[i].buf;
                binds[i].buffer_length = MAXDATALEN;
        }

        if (mysql_stmt_bind_result(stmt, binds) != 0)
                errx(1, "mysql_stmt_bind_result");

        if (mysql_stmt_execute(stmt) != 0)
                errx(1, "mysql_stmt_execute");

        while (true) {
                rc = mysql_stmt_fetch(stmt);

                if (rc == 0) {
                        if (!silent) {
                                for (int i = 0; i < column_count; i++) {
                                        if (i != 0)
                                                printf(" ");
                                        printf("%s", columns[i].buf);
                                }
                                printf("\n");
                        }
                        count++;
                        continue;
                }

                if (rc == MYSQL_NO_DATA)
                        break;

                assert(rc == MYSQL_DATA_TRUNCATED);
                errx(1, "mysql_stmt_fetch: %s", mysql_stmt_error(stmt));
        }

        printf("count: %u\n", count);

        return 0;
}
